#!/usr/bin/env python3
'''
LTE Daemon with autoswitch feature - designed for Unipi PLCs
    Created on 3 Jun 2019 - version 0.26
@author: martyy <triska@unipi.technology>
'''

import os, sys, stat, traceback, string, atexit
import logging, logging.handlers
import re
import time
from enum import Enum
import serial
import signal

try:  # For python2/3 compat
    import ConfigParser
except:
    import configparser as ConfigParser
import argparse

try:
    import exceptions
except:
    pass
import datetime
import fcntl
import struct
import random
import subprocess
import hashlib
from time import sleep

# Default values, can be overridden in the .conf file
DEFAULT_LOGGING_LEVEL = "INFO"
DEFAULT_LOG_PATH = None  # Log path is none -> use syslog instead of an user log file
DEFAULT_LOGFILE_SIZE = 131072  # 128 kiB
RESCUE_THRESHOLD = 350  # Not-success ping to modem restart

# Global definitions
DAEMON_VERSION = "0.46"  # _CIREPL_VERSION_
MODEM_PORT = "/dev/ttyLTE_MNG"  # Symlink to ttyACM0
DEFAULT_MODEM_PORT = "/dev/ttyLTE_MNG"  # Symlink to ttyACM0
CONFILE_PATH = "/etc/unipi/lte.conf"
PIDFILE_PATH = "/run/unipilte.pid"
NIC_CONF_PATH_INTERFACES = "/etc/network/interfaces.d/"
NIC_CONF_PATH_NETWORKD = "/etc/systemd/network/"
PLATFORM_SPEC = {"Axon": {"restart_pin": 18, "usb_port": "3-1:1.0",
                          'leds': {'LED_SIM': "/run/unipi-plc/by-sys/lte.1/led.int.r",
                                   'LED_NET': "/run/unipi-plc/by-sys/lte.1/led.ner.r",
                                   'LED_SIG': "/run/unipi-plc/by-sys/lte.1/led.sig.g",
                                   'LED_INT': "/run/unipi-plc/by-sys/lte.1/led.int.g"
                                   }
                          },
                 "Patron": {"pwrkey_pin": 149, "reset_pin": 84, "usb_port": "1-1:1.0",
                            'leds': {'sim': "/sys/class/leds/unipi:green:sysled-x1",
                                     'net': "/sys/class/leds/unipi:green:sysled-x2",
                                     'sig': "/sys/class/leds/unipi:green:sysled-x3",
                                     'int': "/sys/class/leds/unipi:green:sysled-x4"
                                     }
                            },
                 "ID-LTE-1": {'rfkill_enable' : "/run/unipi-plc/by-sys/lte.1/rfkill",
                              'rfkill_reset' : "/run/unipi-plc/by-sys/lte.1/rfreset", 
                              'leds': {'sim': "!/run/unipi-plc/by-sys/lte.1/led.int.r",
                                       'net': "!/run/unipi-plc/by-sys/lte.1/led.sig.r",
                                       'sig': "/run/unipi-plc/by-sys/lte.1/led.sig.g",
                                       'int': "/run/unipi-plc/by-sys/lte.1/led.int.g"
                                       }
                          }
                 }

# Indicators moved to sysleds (requires unipi-kernel-modules or neuron-kernel version >= 1.42)
# And appropriate firmware (version >=5.50)

LED_SIM_PATH = "/sys/class/leds/unipi:green:sysled-x0"
LED_NET_PATH = "/sys/class/leds/unipi:green:sysled-x1"
LED_SIG_PATH = "/sys/class/leds/unipi:green:sysled-x2"
LED_INT_PATH = "/sys/class/leds/unipi:green:sysled-x3"

# Network states
WAN_STATE_INIT = 0
WAN_STATE_LTE_ON = 1
WAN_STATE_LTE_OFF = 2
WAN_STATE_ETH_ON = 3
WAN_STATE_LTE_ON_ETH_OFF = 4
WAN_STATE_LTE_OFF_ETH_ON = 5
WAN_STATE_LTE_ON_ETH_ON = 6
WAN_STATE_LTE_OFF_ETH_OFF = 7
WAN_STATE_ETH_OFF = 8
WAN_STATE_MONITOR_OFF = 9

# Specific error codes (no all known errors listed)
ERROR_CODE_UNSPEC = 0
ERROR_CODE_NO_SIM = 1
ERROR_CODE_SIM_ERROR = 2
ERROR_CODE_WRONG_PIN = 3

# Actions usually performed in non-standard cases
ACTION_NOTHING = 0
ACTION_CONTINUE = 1
ACTION_REINIT = 2
ACTION_SOFTRESTART = 3
ACTION_HARDRESTART = 4
ACTION_SUSPEND = 5
ACTION_HALT = 6


# Inherited exceptions
class ModemCreateInstanceException(Exception):
    pass


class TimeoutException(Exception):
    pass


class ConnectionError(Exception):
    pass


class ModemErrorException(Exception):
    pass


class LEDErrorException(Exception):
    pass


class NewSettinngs(Exception):
    pass


# Global variables
current_config = {}
logger = None


# Graceful exit with log event generated
def sig_handler(signum, frame):
    logger = logging.getLogger('unipilte')
    logger.info("Exiting after received signal {}".format(signum))
    led = LedServer(PLATFORM_SPEC[ModemFactory.get_platform()[1]]["leds"])
    led.all_off()
    Stats.set_default_values()
    os.unlink(PIDFILE_PATH)
    os._exit(0)


# Some of the network providers can assign WAN IP from the local range(s).
# In the case of IP conflict on both subnets, changes the appropriate IPs.
def resolve_subnets(ip_wan, ip_private, subnet):
    v1_ip = "192.168.101.1"  # The first variant
    v1_dhcp = "192.168.101.100"  # The first variant DHCP range
    v2_ip = "10.11.12.1"  # Second variant
    v2_dhcp = "10.11.12.100"  # Second variant DHCP range
    new_subnet = "255.255.255.0"

    def ip_to_binary(ip):  # Helper function converting string representation of given ip to binary integer
        octet_list_int = ip.split(".")
        octet_list_bin = [format(int(i), '08b') for i in octet_list_int]
        binary = ("").join(octet_list_bin)
        return binary

    subnet_wan = int(ip_to_binary(ip_wan), 2) & int(ip_to_binary(subnet), 2)  # Bitwise AND of IP and SUBNET
    subnet_private = int(ip_to_binary(ip_private), 2) & int(ip_to_binary(subnet), 2)  # Bitwise AND of IP and SUBNET

    logger.debug("WAN IP {}, Private IP {}; {} ".format(ip_wan, ip_private,
                                                        "CONFLICT" if subnet_private == subnet_wan else "NON-CONFLICT"))

    if subnet_wan == subnet_private:  # According to the doc from UBlox, ip range of LAN and WAN must be different
        logger.info("Resolving IP conflict")
        if ip_private == v1_ip:
            return {'ip': v2_ip, 'dhcp': v2_dhcp, 'mask': new_subnet}
        else:
            return {'ip': v1_ip, 'dhcp': v1_dhcp, 'mask': new_subnet}

    return None

# Get path of the interface config file used
def get_lte_interface_confile_path(iface):
    global NIC_CONF_PATH_INTERFACES
    global NIC_CONF_PATH_NETWORKD
    
    if os.path.isfile(NIC_CONF_PATH_INTERFACES + "interfaces"): # TODO Use method
        if logger is not None:
            logger.debug("Debian interfaces are used as network manager.".format(iface))
        return NIC_CONF_PATH_INTERFACES + iface
    else:
        if logger is not None:
            logger.debug("Systemd-networkd are used as network manager.".format(iface)) # TODO - check netwd running
        return NIC_CONF_PATH_NETWORKD + "11-unipi-lte.network"

# Checks if config file (depends on network manager used) exists. If no, creates it.
def check_iface_config_file(iface):

    current_nic_conf_path = get_lte_interface_confile_path(iface)

    if "systemd" in current_nic_conf_path:
        IFACE_FILE_CONTENT = "[Match]\nName={}\n\n[Network]\nDHCP=yes\n".format(iface)
    else:
        IFACE_FILE_CONTENT = "allow-hotplug {0}\niface {0} inet dhcp\n".format(iface)

    if not os.path.isfile(current_nic_conf_path) or IFACE_FILE_CONTENT not in open(current_nic_conf_path).read():

        file = open(current_nic_conf_path, 'w+')
        file.write(IFACE_FILE_CONTENT)
        file.close()
        logger.info("Created/corrected {} iface file".format(iface))
        subprocess.call("/bin/networkctl reload", shell=True)
    else:
        logger.debug("Interface {} iface file is OK and up-to date.".format(iface))

# Get LTE NIC name and creates appropriate config file if necessary
def get_lte_iface_name(nic_id, create_config_file=True, fast=False):
    iface = None
    ifaces_path = '/sys/class/net'

    def look_for_nic():  # Helper function iterating over directory and looking for NIC on desired USB port (e.g. returns eth1 or wwan0)
        for filename in os.listdir(ifaces_path):
            path = os.readlink(ifaces_path + "/" + filename)
            if nic_id in path:
                return filename
        return None


    if ":" in nic_id: # Get nic name from USB id
        for i in range(3):
            iface = look_for_nic()
            if iface:
                break
            if fast is False:
                sleep(3)
    else:
        with open(nic_id, "r") as f:
            iface = f.read()

    if not iface:
        return None

    if create_config_file:
        check_iface_config_file(iface)

    return iface

# Delete file previously created by check_iface_config_file. Used in purge/remove process.
def remove_iface_confile(iface_name):

    if iface_name is None:
        print("No LTE interface found, nothing to remove...")
        return

    iface_path = get_lte_interface_confile_path(iface_name)

    if os.path.exists(iface_path):

        NetworkManagerDriver.set_down(iface_name, force = True)
        print("Removing network interface " + iface_name)
        os.remove(iface_path)
    else:
        print("No interface config found, nothing to remove...")

# Check whether the config file has been changed since the last signature
# If so, signtarure is re-generated (also in case the signature not found)
def was_confile_changed():
    global logger

    changed = False
    with open(CONFILE_PATH, "r+") as f:
        data = f.read().splitlines()
        m = hashlib.md5()
        for i in range(0, len(data) - 1): m.update(data[i].encode())
        print(m.hexdigest())
        if data[-1].strip(" #;\r\n\t") != m.hexdigest():
            changed = True
            signature = "; " + m.hexdigest()

            if ';' in data[-1][0:3]:
                data[-1] = signature
            else:  # Create new signature - used during migration from non-signed confile
                data.append("; Config file signature - do not modify it nor add any lines or whitespaces!")
                data.append(signature)

            f.seek(0)
            f.write("\n".join(data))

    return changed


# Parse of config file with int/bool data types conversion and default values
def getFromConfile(config, section, option, default, boolean=False, integer=False):
    global logger
    try:
        value = config.get(section, option)
        if boolean == True:
            value = value.lower() in ("yes", "true", "t", "1")
        elif integer == True:
            value = int(value)
    except ValueError:
        if logger is not None:
            logger.warning(
                "Cannot parse integer parameter {}:{}, using default value: {}".format(section, option, default))
        return default
    except:
        if logger is not None:
            logger.debug("Cannot parse parameter {}:{}, using default value: {}".format(section, option, default))
        return default

    return value



class NetworkManagerDriver(object):
    
    class NMType(Enum):
        INTERFACES = 1
        SYSTEMD = 2
    
    Color = Enum('mmtype', ['networking', 'systemd'])

    def __set_up_networking(iface):
        subprocess.call("/sbin/ip link set dev {} up".format(iface), shell=True)
        subprocess.call("/sbin/ifup " + iface, shell=True)  # Can take longer
    
    def __set_down_networking(iface, force):
        subprocess.call("/sbin/ip link set dev {} down".format(iface), shell=True)
        subprocess.call("/sbin/ifdown {} " .format("--force " if force is True else "") + iface, shell=True)  # Can take longer
    
    def __set_up_systemd(iface):
        subprocess.call("/sbin/ip link set dev {} up".format(iface), shell=True)
        subprocess.call("/bin/networkctl up " + iface, shell=True)
    
    def __set_down_systemd(iface, force):
        subprocess.call("/sbin/ip link set dev {} down".format(iface), shell=True)
        subprocess.call("/bin/networkctl down " + iface, shell=True)
        if force is True:
            subprocess.call("/bin/networkctl reload", shell=True)

    __up_function :callable = None
    __down_function : callable = None
    
    @classmethod
    def set_up(cls, iface):
        cls.__up_function(iface)
    
    @classmethod
    def set_down(cls,iface, force = False):
        cls.__down_function(iface, force)
    
 
    @classmethod
    def init(cls):
        if NetworkManagerDriver.get_network_manager_type() == NetworkManagerDriver.NMType.INTERFACES:
            cls.__up_function = NetworkManagerDriver.__set_up_networking
            cls.__down_function = set_down_networking.__set_down_networking
        elif  NetworkManagerDriver.get_network_manager_type() == NetworkManagerDriver.NMType.SYSTEMD:
            cls.__up_function = NetworkManagerDriver.__set_up_systemd
            cls.__down_function = NetworkManagerDriver.__set_down_systemd
    
    @staticmethod
    def get_network_manager_type():
        if os.path.isfile(NIC_CONF_PATH_INTERFACES + "interfaces") :
            return NetworkManagerDriver.NMType.INTERFACES
        else :
            return NetworkManagerDriver.NMType.SYSTEMD

# Collects and computes statistical data related to cellular network quality and WAN network availability
class Stats(object):
    SYSFS_PATH = "/run/unipi_lte/"

    SYSFS_RSSI_FILE = "rssi"
    SYSFS_SIGQUAL_FILE = "sigqual"
    SYSFS_MODE_FILE = "mode"
    SYSFS_NETTYPE_FILE = "nettype"

    sysfs_files = (SYSFS_RSSI_FILE,
                   SYSFS_SIGQUAL_FILE,
                   SYSFS_MODE_FILE,
                   SYSFS_NETTYPE_FILE)

    def __init__(self, period=120, ping_active=False):
        self.period = period
        self.states = set()
        self.ping_active = ping_active
        self.reset()

        if not os.path.exists(Stats.SYSFS_PATH):
            os.mkdir(Stats.SYSFS_PATH)

        for filename in Stats.sysfs_files:
            with open(Stats.SYSFS_PATH + filename, "w") as file:
                file.write("-1")

    @staticmethod
    def set_default_values():
        for filename in Stats.sysfs_files:
            with open(Stats.SYSFS_PATH + filename, "w") as file:
                file.write("-1")

    def __del__(self):
        Stats.set_default_values()

    def report_partial(self, rssi, sigqual, nettype):  # Commit often read values only, called from main loop

        try:
            rssi = int(rssi)
            sigqual = int(sigqual)
        except:
            return

        self.nettypes.add(nettype)

        if rssi == 99:  # Not valid value
            return

        self.cnt = self.cnt + 1
        if self.rssi_max < rssi:
            self.rssi_max = rssi
        elif self.rssi_min > rssi:
            self.rssi_min = rssi
        self.rssi_sum += rssi

        self.sigqual = sigqual

        with open(stats.SYSFS_PATH + Stats.SYSFS_NETTYPE_FILE, "w") as file:
            file.write(str(nettype))

        # Mirror values to sysfs

        with open(stats.SYSFS_PATH + Stats.SYSFS_RSSI_FILE, "w") as file:
            file.write(str(rssi))

        with open(stats.SYSFS_PATH + Stats.SYSFS_SIGQUAL_FILE, "w") as file:
            file.write(str(sigqual))

    def report_metrics(self, rssi, sigqual, wan_status, nettype,
                       ping_result=None):  # Commit all values, called form main loop

        self.report_partial(rssi, sigqual, nettype)

        self.states.add(wan_status)

        if ping_result is not None:
            if ping_result is not True:
                self.pings_failed += 1

            self.pings_cnt += 1

        with open(stats.SYSFS_PATH + Stats.SYSFS_MODE_FILE, "w") as file:
            file.write(str(wan_status))

    def reset(self):
        self.cnt = 0
        self.rssi_sum = 0
        self.rssi_min = 100
        self.rssi_max = 0
        self.pings_cnt = 0
        self.pings_failed = 0
        self.last_summary = time.time()
        self.sigqual = 0
        self.nettypes = set()
        self.states = set()

    def get_summmary(self, do_reset=True):  # Get report output in text form (e.g. for periodical storing to a log file)

        if self.ping_active and self.pings_cnt != 0:
            sum_str = "RSSI: (MIN {}, MAX {}, AVG {}), SIGQUAL: {}, NET-TYPES: {}, USED_MODES: {} PING_COUNT: {} PING_SUCCESS_RATIO: {:0.2f}".format(
                self.rssi_min,
                self.rssi_max,
                int(self.rssi_sum / self.cnt),
                self.sigqual,
                list(self.nettypes),
                self.states,
                self.pings_cnt,
                ((self.pings_cnt - self.pings_failed) / self.pings_cnt) * 100)
        elif self.cnt > 0:
            sum_str = "RSSI: (MIN {}, MAX {}, AVG {}) SIGQUAL: {} NET-TYPES: {} ".format(
                self.rssi_min,
                self.rssi_max,
                int(self.rssi_sum / self.cnt),
                self.sigqual,
                list(self.nettypes))

        if do_reset: self.reset()
        return sum_str

    def is_time_fullfilled(self):  # Returns true if it is time to generate stats into the log file
        if not self.period:
            return False
        return (self.last_summary + self.period) < time.time()


# Provides ping and autoswitch functionality
class ConnectionGuard(object):
    GW_MASK_LTE = 1
    GW_MASK_ETH = 2

    def isOnline(self):  # True if the unit has access to the WAN network (found by ping)

        if self.wan_state not in (WAN_STATE_LTE_OFF, WAN_STATE_LTE_OFF_ETH_OFF, WAN_STATE_INIT):
            return False
        return True

    def __init__(self):

        global current_config
        self.local_lte_ip = None
        self.current_eth_metric = 0
        self.current_lte_metric = 0
        self.gw_bitmask = 0
        self.rtable_valid = True
        self.autoswitch_was_active = False
        self.reset_state()

    def reset_state(self):
        self.wan_state = WAN_STATE_INIT if current_config['ping']['active'] is True else WAN_STATE_MONITOR_OFF
        self.fail_counterA = 0
        self.fail_counterB = 0
        self.rescue_counter = 0

    # Detects if the main interface is not available (e.g. cable is off)
    # Reads and compares metrics of the intefaces (both main and backup)
    def __read_default_gateway(self):
        global current_config

        self.current_eth_metric = -1
        self.current_lte_metric = -1
        gw_bitmask = 0
        routecount = 0

        STATE_PATH = "/sys/class/net/{}/operstate"
        upstrings = ("up", "unknown")

        try:
            with open(STATE_PATH.format(current_config['ping']['eth_iface']), "r") as f:
                res = f.read()
                if any(x in res for x in upstrings):
                    gw_bitmask |= self.GW_MASK_ETH
        except:
            pass

        try:
            with open(STATE_PATH.format(current_config['lte_iface']), "r") as f:
                res = f.read()
                if any(x in res for x in upstrings):
                    gw_bitmask |= self.GW_MASK_LTE
        except:
            pass

        with open("/proc/net/route") as fh:
            for line in fh:
                routecount = routecount + 1
                fields = line.strip().split()  # Default gateway ha
                if (fields[1] == '00000000') and (fields[0] == current_config['ping']['eth_iface']):
                    self.current_eth_metric = int(fields[6])
                elif (fields[1] == '00000000') and (fields[0] == current_config['lte_iface']):
                    self.current_lte_metric = int(fields[6])

                    if self.local_lte_ip is None:
                        buffer = bytearray(4)
                        struct.pack_into("<L", buffer, 0, int(fields[2], 16))
                        self.local_lte_ip = ".".join(str(c) for c in buffer)
                        logger.debug("Assigned local IP " + self.local_lte_ip)

                        if self.local_lte_ip == "0.0.0.0":
                            self.local_lte_ip = None
                        # logger.debug("GW IP {}".format(self.local_lte_ip))

        if self.current_eth_metric == -1:
            logger.debug("Cannot obtain metric for {} (is up: {})".format(current_config['ping']['eth_iface'],
                                                                          gw_bitmask & self.GW_MASK_ETH))
            gw_bitmask &= ~(self.GW_MASK_ETH)

        if self.current_lte_metric == -1:
            logger.debug("Cannot obtain metric for {} (is up: {})".format(current_config['lte_iface'],
                                                                          gw_bitmask & self.GW_MASK_LTE))
            gw_bitmask &= ~(self.GW_MASK_LTE)

        is_rtable_valid = (True if (routecount > 2) else False)
        # is_current_iface_main = (True if ((main_metric < backup_metric) and (gw_bitmask & 1)) else False)

        logger.debug(
            "Current iface metrics main_iface:{} backup_iface:{} Local LTE ip {}, routing items {}, gw_bitmask {}".format(
                self.current_eth_metric,
                self.current_lte_metric, self.local_lte_ip,
                routecount, gw_bitmask))

        return (gw_bitmask, is_rtable_valid)

    def keep_iface_on(self):

        if current_config['ping']['autoswitch'] is False:
            self.gw_bitmask, is_rtable_valid = self.__read_default_gateway()

        if (not (self.gw_bitmask & self.GW_MASK_LTE)):
            logger.warning(
                "UP {} LTE default route is MISSING, restart the iface...".format(current_config['lte_iface']))
            if self.try_interface_up(current_config['lte_iface'], down_firstly=True) is False:
                raise ConnectionError()

    @staticmethod
    def try_interface_up(iface, down_firstly=False):
        if down_firstly is True:
            NetworkManagerDriver.set_down(iface, force=True)

        def has_ip_assigned():
            ret = subprocess.check_output("ip addr show " + iface, shell=True)
            logger.debug(ret)
            if b"inet " in ret and b"DOWN" not in ret and b"169.254" not in ret:
                return True
            return False

        if has_ip_assigned() is True and down_firstly is False:
            return True

        for i in range(2):  # Wait max 18 sec for DHCP IP address assignment
            NetworkManagerDriver.set_up(iface)
            for j in range(8):
                if has_ip_assigned() is True: return
                logger.debug(
                    "Waiting for IP assign...remain {}:{} down firstly {}".format(str(2 - i), str(8 - j), down_firstly))
                time.sleep(8)

        if down_firstly is False:
            logger.warning("Cannot obtain IP address from LTEs DHCP server...iface restart")
            return ConnectionGuard.try_interface_up(iface, down_firstly=True)

        else:
            logger.error("Cannot obtain IP address from LTEs DHCP server...")
            return False

    @staticmethod
    def update_resolver_file(lte_as_dns, dns_ip=None):

        global logger

        if NetworkManagerDriver.get_network_manager_type() == NetworkManagerDriver.NMType.SYSTEMD :
            if lte_as_dns is True:
                if logger is not None:
                    logger.debug("Changing domain for DNS resolving to LTE iface ")
                ret = subprocess.check_output("resolvectl domain {} ~.".format(current_config['lte_iface']) , shell=True)
            else:
                if logger is not None:
                    logger.debug("Changing domain for DNS resolving to main iface ")
                ret = subprocess.check_output("resolvectl domain {} \"\"".format(current_config['lte_iface']) , shell=True)

            return

        FILEPATH = "/etc/resolvconf/resolv.conf.d/head"
        LINE = "nameserver {}\n".format(dns_ip)  # Toto vznikne
        data = None
        changed = False

        def remove_current_records(data):
            data = str(data).splitlines(True)
            new_content = ""
            line_num = 0
            changed = False
            while line_num < len(data):
                if "unipi-lte" in data[line_num]:  # skip
                    line_num = line_num + 2
                    changed = True
                    continue
                new_content += (data[line_num])
                line_num = line_num + 1

            with open(FILEPATH, "w") as f:
                f.write(new_content)

            return changed

        try:  # If the config file does not exists
            with open(FILEPATH, "r") as f:
                data = f.read()
        except Exception as e:
            if logger is not None:
                logger.warning("Cannot change DNS resolver file: " + str(e))
            return

        if LINE in data:
            newdata = re.sub(r"\#?({})".format(LINE), r"{}\1".format("" if lte_as_dns is True else "#"), data)
            if newdata != data:
                with open(FILEPATH, "w") as f:
                    f.write(newdata)
                changed = True

        elif lte_as_dns == True and dns_ip is not None:
            remove_current_records(data)
            MSG = "#This line has been added by unipi-lte daemon\n{}".format(LINE)
            with open(FILEPATH, "a") as f:
                f.write(MSG)
            changed = True

        else:  # Remove records
            changed = remove_current_records(data)

        if changed is not False:
            status = subprocess.call("resolvconf " + "-u", shell=True)

    def __switch_default_iface(self, toBackup):  # Switches network ifaces (change their metrics)
        global current_config

        new_metric_shift = 0
        # DEF Change metric - lokalni funkce

        if toBackup and (self.gw_bitmask & self.GW_MASK_LTE):

            logger.debug("Switching to LTE interface {}".format(current_config['lte_iface']))

            if not (self.gw_bitmask & self.GW_MASK_LTE):
                return

            if self.current_lte_metric + 1 == self.current_eth_metric:
                new_metric_shift = 1

            new_metric = self.current_eth_metric - 1 - new_metric_shift
            if new_metric <= 0:
                new_metric = 1

            logger.debug(
                "ip route add default via {} dev {} metric {}".format(self.local_lte_ip, current_config['lte_iface'],
                                                                      new_metric))
            logger.debug(
                "ip route delete default dev {} metric {}".format(current_config['lte_iface'], self.current_lte_metric))

            os.system(
                "ip route add default via {} dev {} metric {}".format(self.local_lte_ip, current_config['lte_iface'],
                                                                      new_metric))
            os.system(
                "ip route delete default dev {} metric {}".format(current_config['lte_iface'], self.current_lte_metric))

            self.current_lte_metric = new_metric

            try:
                self.update_resolver_file(lte_as_dns=True, dns_ip=self.local_lte_ip)
            except Exception as E:
                logger.critical("Failure: " + str(E))

        elif self.gw_bitmask & self.GW_MASK_ETH:

            logger.debug("Switching to main interface {}".format(current_config['ping']['eth_iface']))

            if self.current_lte_metric - 1 == self.current_eth_metric:
                new_metric_shift = 1

            new_metric = self.current_eth_metric + 1 + new_metric_shift

            logger.debug(
                "ip route add default via {} dev {} metric {}".format(self.local_lte_ip, current_config['lte_iface'],
                                                                      new_metric))
            logger.debug(
                "ip route delete default dev {} metric {}".format(current_config['lte_iface'], self.current_lte_metric))
            os.system(
                "ip route add default via {} dev {} metric {}".format(self.local_lte_ip, current_config['lte_iface'],
                                                                      new_metric))
            os.system(
                "ip route delete default dev {} metric {}".format(current_config['lte_iface'], self.current_lte_metric))

            self.current_lte_metric = new_metric

            try:
                self.update_resolver_file(lte_as_dns=False, dns_ip=self.local_lte_ip)
            except Exception as E:
                logger.critical("Failure: " + str(E))

        logger.debug("Autoswitch complete")

    def get_wan_status(self):  # Get current status of the connection (see WAN_STATE_xxxx constants)
        return self.wan_state

    def get_rtable_valid(self):  # Returns true if routing table contains at least one valid record
        return self.rtable_valid

    def __autoswitch(self, lte_status):  # Autoswitch core called from ping function (if enabled in the config file)

        global current_config

        LTE_UNCHANGED = 0
        LTE_OK = 1
        LTE_OFFLINE = 2

        ETH_UNCAHNGED = 0
        ETH_GOES_OFFLINE = 1
        ETH_GOES_ONLINE = 2
        eth_status = ETH_UNCAHNGED

        response_b = 1  # Ping failed as default

        self.gw_bitmask, self.rtable_valid = self.__read_default_gateway()

        logger.debug("GW MASK: " + str(self.gw_bitmask))

        if not self.rtable_valid:  # Do not perform autoswitch if routing table is not valid (typically after startup)
            logger.debug("Routing table is not (yet) valid, autoswitch skipping...")
            return

        elif self.gw_bitmask != (self.GW_MASK_ETH | self.GW_MASK_LTE) and self.autoswitch_was_active is False:
            logger.debug("No more default gateways, autoswitch skipping...")
            if lte_status == LTE_OK:
                if self.wan_state != WAN_STATE_LTE_ON_ETH_OFF:
                    self.__switch_default_iface(toBackup=True)
                self.wan_state = WAN_STATE_LTE_ON_ETH_OFF
            else:
                self.wan_state = WAN_STATE_LTE_OFF_ETH_OFF
            return

        else:
            self.autoswitch_was_active = True
            logger.debug("Autoswitch ACTIVE ")

        # -----------------------------------------------------------------------------------------------------------

        if self.gw_bitmask & self.GW_MASK_ETH:  # If the main iface is present, ping it
            response_b = os.system("ping -c 1 -W {0} -I {1} -s {2} {3} >/dev/null".format(
                current_config['ping']['timeout'],
                current_config['ping']['eth_iface'],
                "32", current_config['ping']['ip_eth']))
        elif self.wan_state in {WAN_STATE_LTE_OFF_ETH_ON, WAN_STATE_LTE_ON_ETH_ON,
                                WAN_STATE_INIT}:  # If main interface has been disconnected, switch to backup immediately
            self.fail_counterB = current_config['ping']['threshold'] + 1
            logger.info("Main GW is not present ! Switching to backup iface {} immediately...".format(
                current_config['lte_iface']))

        if response_b and self.wan_state not in {WAN_STATE_ETH_OFF, WAN_STATE_LTE_OFF_ETH_OFF,
                                                 WAN_STATE_LTE_ON_ETH_OFF}:  # Ping on main interface failed and ETH is ON

            if self.fail_counterB >= current_config['ping']['threshold']:  # If counter overflows, ETH ON -> ETH OFF
                eth_status = ETH_GOES_OFFLINE
                self.__switch_default_iface(toBackup=True)  # Switch to backup
                self.gw_bitmask, self.rtable_valid = self.__read_default_gateway()  # Refresh GW info
                if self.fail_counterB == current_config['ping']['threshold']:
                    logger.info("Ping failure threshold reached {} on main {}.Switching to backup {}...".format(
                        current_config['ping']['threshold'],
                        current_config['ping']['eth_iface'],
                        current_config['lte_iface']))

            else:
                self.fail_counterB = self.fail_counterB + 1

            logger.debug("Ping failed on {} no. {}; Threshold: {}".format(
                current_config['ping']['eth_iface'],
                self.fail_counterB, current_config['ping']['threshold']))

        elif (not response_b) and self.wan_state not in {WAN_STATE_ETH_ON, WAN_STATE_LTE_OFF_ETH_ON,
                                                         WAN_STATE_LTE_ON_ETH_ON}:  # Ping OK, switch back to main iface

            if self.wan_state != WAN_STATE_INIT:
                logger.info("Main interface {} becomes available, switching to it".format(
                    current_config['ping']['eth_iface']))
                self.__switch_default_iface(toBackup=False)  # Switch to main
                self.gw_bitmask, self.rtable_valid = self.__read_default_gateway()  # Refresh GW info

            eth_status = ETH_GOES_ONLINE
            logger.debug("ETH Ping OK")

        else:
            if not response_b:
                logger.debug("ETH Ping OK")
            self.fail_counterB = 0

        for i in range(2):  # Resolve current state variable
            if eth_status == ETH_GOES_ONLINE:
                if self.wan_state in (WAN_STATE_INIT, WAN_STATE_ETH_OFF):
                    self.wan_state = WAN_STATE_ETH_ON
                elif self.wan_state in (WAN_STATE_LTE_ON, WAN_STATE_LTE_ON_ETH_OFF):
                    self.wan_state = WAN_STATE_LTE_ON_ETH_ON
                elif self.wan_state in (WAN_STATE_LTE_OFF, WAN_STATE_LTE_OFF_ETH_OFF):
                    self.wan_state = WAN_STATE_LTE_OFF_ETH_ON

            elif eth_status == ETH_GOES_OFFLINE:
                if self.wan_state in (WAN_STATE_INIT, WAN_STATE_ETH_ON):
                    self.wan_state = WAN_STATE_ETH_OFF
                elif self.wan_state in (WAN_STATE_LTE_ON, WAN_STATE_LTE_ON_ETH_ON):
                    self.wan_state = WAN_STATE_LTE_ON_ETH_OFF
                elif self.wan_state in (WAN_STATE_LTE_OFF, WAN_STATE_LTE_OFF_ETH_ON):
                    self.wan_state = WAN_STATE_LTE_OFF_ETH_OFF

            if lte_status == LTE_OK:
                if self.wan_state in (WAN_STATE_INIT, WAN_STATE_LTE_OFF):
                    self.wan_state = WAN_STATE_LTE_ON
                elif self.wan_state in (WAN_STATE_ETH_ON, WAN_STATE_LTE_OFF_ETH_ON):
                    self.wan_state = WAN_STATE_LTE_ON_ETH_ON
                elif self.wan_state in (WAN_STATE_ETH_OFF, WAN_STATE_LTE_OFF_ETH_OFF):
                    self.wan_state = WAN_STATE_LTE_ON_ETH_OFF

            elif lte_status == LTE_OFFLINE:
                if self.wan_state in (WAN_STATE_INIT, WAN_STATE_LTE_ON):
                    self.wan_state = WAN_STATE_LTE_OFF
                elif self.wan_state in (WAN_STATE_ETH_ON, WAN_STATE_LTE_ON_ETH_ON):
                    self.wan_state = WAN_STATE_LTE_OFF_ETH_ON
                elif self.wan_state in (WAN_STATE_ETH_OFF, WAN_STATE_LTE_ON_ETH_OFF):
                    self.wan_state = WAN_STATE_LTE_OFF_ETH_OFF

        if self.gw_bitmask == (self.GW_MASK_ETH | self.GW_MASK_LTE):

            if (self.wan_state == WAN_STATE_LTE_ON_ETH_OFF) and (
                    self.current_lte_metric >= self.current_eth_metric):  # In backup mode
                logger.warning(
                    "Routing table metric mismatch A (changed by other application?), re-setting... {} {}".format(
                        self.current_lte_metric, self.current_eth_metric))
                self.__switch_default_iface(toBackup=True)

            elif (self.wan_state in (WAN_STATE_LTE_OFF_ETH_ON, WAN_STATE_LTE_ON_ETH_ON)) and (
                    self.current_eth_metric >= self.current_lte_metric):
                logger.warning("Routing table metric mismatch B (changed by other application?), re-setting...")
                self.__switch_default_iface(toBackup=False)

        return True
        # return not (bool(response_main and response_backup))

    def ping(self):  # Checks the WAN connection, set state varaibles(s) and handle autoswitch stuff if enabled
        global current_config

        LTE_UNCHANGED = 0
        LTE_OK = 1
        LTE_OFFLINE = 2
        lte_status = LTE_UNCHANGED

        logger.debug("ping -c 1 -W {0} -I {1} -s {2} {3} >/dev/null".format(current_config['ping']['timeout'],
                                                                            current_config['lte_iface'], "32",
                                                                            current_config['ping']['ip_lte']))
        # Ping on the backup iface (LTE)
        response_lte = os.system("ping -c 1 -W {0} -I {1} -s {2} {3} >/dev/null".format(
            current_config['ping']['timeout'],
            current_config['lte_iface'],
            "32",
            current_config['ping']['ip_lte']))

        if response_lte:  # Ping not successful
            if self.rescue_counter >= 0:
                logger.debug("RESCUE-current: {}\nRESCUE-max: {}".format(self.rescue_counter, RESCUE_THRESHOLD))
                if self.rescue_counter > RESCUE_THRESHOLD:
                    self.rescue_counter = 0
                    raise ConnectionError

                self.rescue_counter = self.rescue_counter + 1

            logger.debug(
                "Ping failed GSM no. {}; Threshold: {}".format(self.fail_counterA, current_config['ping']['threshold']))
            if self.fail_counterA >= current_config['ping']['threshold']:

                if self.wan_state in {WAN_STATE_INIT, WAN_STATE_ETH_ON}:
                    logger.info("LTE: Cannot get to online state {}".format(self.wan_state))
                elif self.wan_state in {WAN_STATE_LTE_ON, WAN_STATE_LTE_ON_ETH_OFF, WAN_STATE_LTE_ON_ETH_ON}:
                    logger.info("LTE: Goes offline (previously online)")

                lte_status = LTE_OFFLINE
            else:
                self.fail_counterA = self.fail_counterA + 1
        else:  # Ping was OK
            self.fail_counterA = 0
            self.rescue_counter = 0
            if self.wan_state in {WAN_STATE_LTE_OFF, WAN_STATE_LTE_OFF_ETH_OFF, WAN_STATE_LTE_OFF_ETH_ON}:
                logger.info("LTE: Goes online (previously offline) {}".format(self.wan_state))
            lte_status = LTE_OK
            logger.debug("LTE Ping OK")

        if current_config['ping']['autoswitch'] is True:  # Autoswitch feature is active and LTE status is resolved
            self.__autoswitch(lte_status)
        else:  # Autoswitch disabled, only backup iface periodical probing is active
            if lte_status == LTE_OK:
                self.wan_state = WAN_STATE_LTE_ON
            elif lte_status == LTE_OFFLINE:
                self.wan_state = WAN_STATE_LTE_OFF

        logger.debug("Current WAN state {}".format(self.wan_state))

        return True if lte_status == LTE_OK else False


def do_action(modem, action,
              errcode=ERROR_CODE_UNSPEC):  # Perform desired action (as a reaction to different error conditions)

    reinit = False

    if action == ACTION_SOFTRESTART:
        logger.info("Modem soft-restart")
        try:
            modem.restart(soft=True)
        except Exception as e:
            logger.info("Modem soft-restart failed, performing HARD-restart: " + str(e))
            do_action(modem, ACTION_HARDRESTART)  # Hard restart as a fallback to the SOFT one (via AT command)

        reinit = True  # Re-init modem

    elif action == ACTION_HARDRESTART:
        logger.info("Modem hard-restart caused by an error")
        NetworkManagerDriver.set_down(current_config['lte_iface'], force=True)

        if modem:
            modem.restart_hw()
            del modem

        reinit = True  # Re-init modem

    elif action == ACTION_SUSPEND:
        logger.warning("Daemon suspended...exit in {} sec".format(current_config['suspend_timeout']))
        if current_config['suspend_timeout'] < 0:
            do_action(modem, ACTION_HALT, errcode)
        if modem:
            modem.go_halt_suspend(errcode)
        time.sleep(current_config['suspend_timeout'])
        reinit = True  # Re-init modem

    elif action == ACTION_HALT:
        logger.warning("Daemon halted forever...must be restarted manually {}".format(errcode))
        if modem:
            modem.go_halt_suspend(errcode)
        while True:
            time.sleep(current_config['suspend_timeout'])

    return reinit


class LedServer(object):

    def __init__(self, led_paths):  # Check presence of all leds
        ret: bool = True
        
        self.led_inverted = {}
        self.leds = {}
        print(str(led_paths) )

        for led, path in led_paths.items():
            
            if "!" in path:
                self.led_inverted[led] = True
            else:
                self.led_inverted[led] = False
                
            self.leds[led] = path.strip('!')
            
            ret &= os.path.exists(self.leds[led])

        self.led_states = {"off": "0",
                           "on": "1",
                           "off_inverted" : "1",
                           "on_inverted" : "0",
                           "blink_fast": "150",
                           "blink_slow": "2000"}

        self.led_cache = {}
        for led in self.leds:  # Init da cache
            self.led_cache[led] = 0
            self.led_cache[led + '_duty'] = 0

        if not ret:
            raise LEDErrorException

        self.all_off()


    def set(self, led, state):  # Set state of LTE coupled LEDs

        if self.led_cache[led] == state:
            return

        state = state.lower()
        if state == "on" or state == "off":
            value = open(self.leds[led] + "/trigger", "w")
            value.write(str("none"))
            value.close()

            if self.led_inverted[led] is True:
                state = state + "_inverted"

            value = open(self.leds[led] + "/brightness", "w")
            value.write(str(self.led_states[state]))
            value.close()
        else:  # Blink
            value = open(self.leds[led] + "/trigger", "w")
            value.write(str("timer"))
            value.close()

            value = open(self.leds[led] + "/delay_on", "w")
            value.write(str(self.led_states[state]))
            value.close()

            value = open(self.leds[led] + "/delay_off", "w")
            value.write(str(self.led_states[state]))
            value.close()

        self.led_cache[led] = state

    def duty(self, led, duty):  # Set LED duty cycle (used as a signal quality indicator)
        if self.led_cache[led + '_duty'] == duty and self.led_cache[led] == "duty":
            return

        per = (3000 / 31) * (31 - duty)
        t_on = int(per)  # 0-930
        t_off = int(per)

        value = open(self.leds[led] + "/trigger", "w")
        value.write("timer")
        value.close()

        value = open(self.leds[led] + "/delay_on", "w")
        value.write(str(t_on))
        value.close()

        value = open(self.leds[led] + "/delay_off", "w")
        value.write(str(t_off))
        value.close()

        self.led_cache[led + '_duty'] = duty
        self.led_cache[led] = "duty"

    def all_off(self):  # Switch of all LEDs
        for led in self.leds:
            self.set(led, 'off')

    def all_on(self):  # Switch of all LEDs
        for led in self.leds:
            self.set(led, 'on')

    def all_blink(self, fast=True):
        mode = 'off'
        for led in self.leds:
            self.set(led, mode)
        mode = 'blink_fast' if fast else 'blink_slow'
        for led in self.leds:
            self.set(led, mode)


class LTEModemInterface(object):
    rept_rst_cnt = 0
    last_rst_time = time.time()

    AT_OK_REPLY = b"OK"

    REGEX_CSQ = re.compile(r"\+CSQ:\s*(\d+),(\d+)")
    REGEX_COP = re.compile(r"\+COPS:\s*\d+,\d?,\"(.*)\",(\d?)")

    def __init__(self, nic_usb_port, led_definitions, at_port_path=None, nic_name=None):

        self.at_port_path = DEFAULT_MODEM_PORT if at_port_path is None else at_port_path
        self.nic_id = None
        
        if nic_name is not None:
            self.nic_id = nic_name
        else:
            self.nic_id = nic_usb_port
    
        self.leds = None
        self.iface = None

        if not self.leds:
            self.leds = LedServer(led_definitions)
            self.leds.all_on()

        atexit.register(self.__do_cleanup)

    def init(self, retry_counter=5):

        exc = None
        delay = 8

        retry_cnt = retry_counter / 2

        while retry_cnt > 0:
            try:
                self.__iface_init()
                self.leds.all_off()
                logger.debug("Init OK 1")
                print("Init OK")
                return

            except Exception as E:
                logger.debug(str(E) + " - attempt {}/{}".format(retry_cnt, retry_counter))
                print(E)
                retry_cnt = retry_cnt - 1
                sleep(1)

        self.enable()
        sleep(delay)

        while retry_counter > 0:
            try:
                self.iface = None
                self.__iface_init()
                self.leds.all_off()
                logger.debug("Init OK 2")
                print("Init OK")
                return
            except Exception as E:
                print(E)
                logger.debug(str(E))
                retry_counter = retry_counter - 1
                if retry_counter < 2:
                    logger.info("Modem init retry with HW restart...{} attempts left".format(retry_counter))
                    # self.restart(soft = False)
                    self.restart_hw()
                    delay = 10  # Extend delay to wait for modem boot
                else:
                    logger.info("Modem init retry...{} attempts left".format(retry_counter))
                    self.enable()

                sleep(random.random() * 10 + delay)
                exc = E

                continue

        if self.leds:
            self.leds.all_blink(fast=False)

        raise ModemCreateInstanceException(str(exc))

    def __iface_init(self):
        if not stat.S_ISCHR(os.stat(self.at_port_path).st_mode):  # Check if modem virtual serial interface exists
            raise ModemErrorException("Modem serial port {} is not a valid character device".format(port))
        if hasattr(self.iface, 'open'):
            self.iface.close()

        try:
            self.iface = serial.Serial(self.at_port_path, 115200, timeout=0.3, parity=serial.PARITY_NONE, rtscts=0)
        except Exception as E:
            raise ModemErrorException("Cannot open virtual serial port {}".format(USB_PORT))

        current_config['lte_iface'] = get_lte_iface_name(self.nic_id)
        check_iface_config_file(current_config['lte_iface'])

        if not current_config['lte_iface']:
            raise ModemErrorException("No network interface found on USB port {}:{}".format(port, USB_PORT))

    def query(self, query, bytes_expected=569):
        chars = ""
        retry_cnt = 3

        while retry_cnt > 0:
            try:
                self.iface.flushOutput()  # Flush da buffer
                self.iface.flushInput()  # Flush da buffer
                # time.sleep(0.1)
                self.iface.write(query.encode())  # Write op. is blocking - may take longer !
                chars = self.iface.read(bytes_expected)  # Read with previously defined timeout
            except Exception as e:  # May occur when other app uses modem's serial line
                raise ModemErrorException("Modem I/O error " + str(e))
            # logger.debug("RECEIVED {} bytes: {}".format(len(chars),chars))
            if b"ERROR" in chars:
                raise ModemErrorException(chars)
            if len(chars) > 0:  # No reply received
                return chars

            retry_cnt = retry_cnt - 1
            if len(query) > 4:  # Supress AT\r\n polling during startup sequence
                logger.info("More attempts for " + str(query).strip('\r\n'))

        raise TimeoutException

    def query_extended_timeout(self, query, timeout, bytes_expected=569):

        ret = ""
        timeout_bckp = self.iface.timeout
        self.iface.timeout = timeout
        try:
            ret = self.query(query, bytes_expected)
        finally:
            self.iface.timeout = timeout_bckp
        return ret

    def get_sig_metrics(self):
        ret = self.query("AT+CSQ\r\n")
        res = self.REGEX_CSQ.search(str(ret))
        if res is None or len(res.groups()) != 2:
            logger.info("Cannot parse signal quality, received: {}".format(ret))
            return (99, 0)
        return (int(res.group(1)), int(res.group(2)))

    def get_operator_nettype(self):
        ret = self.query("AT+COPS?\r\n")
        res = self.REGEX_COP.search(str(ret))
        if res is None or len(res.groups()) != 2:
            # logger.info("Cannot parse operator info, received: {}".format(ret))
            return (None, None)

        return (res.group(1), res.group(2))

    def update_led(self, is_registered, rssi, wan_state):

        if is_registered is True:  # Registered in the network
            self.leds.set('net', 'on')
        else:
            self.leds.set('net', 'off')

        if rssi == 99:
            self.leds.set('sig', 'off')
        else:
            self.leds.duty('sig', rssi)

        if current_config['ping']['autoswitch']:
            if wan_state == WAN_STATE_LTE_ON_ETH_ON:  # Connection to the internet (or other WAN network)
                self.leds.set('int', 'on')
            elif wan_state in {WAN_STATE_LTE_ON_ETH_OFF, WAN_STATE_LTE_ON}:
                self.leds.set('int', 'blink_fast')
            elif wan_state in {WAN_STATE_LTE_OFF_ETH_ON, WAN_STATE_ETH_ON}:
                self.leds.set('int', 'blink_slow')
            else:  # OFFLINE
                self.leds.set('int', 'off')
        else:  # No autoswitch
            if wan_state in {WAN_STATE_ETH_ON, WAN_STATE_LTE_ON}:
                self.leds.set('int', 'on')
            else:
                self.leds.set('int', 'off')

    def go_halt_suspend(self, errcode=ERROR_CODE_UNSPEC):

        if errcode in (ERROR_CODE_NO_SIM, ERROR_CODE_SIM_ERROR):
            self.leds.all_off()
            self.leds.set('sim', 'blink_slow')
        elif errcode == ERROR_CODE_WRONG_PIN:
            self.leds.all_off()
            self.leds.set('sim', 'blink_fast')
        else:
            self.leds.all_blink(fast=False)

    def __do_cleanup(self):

        print("Modem ending...")
        if self.iface:
            print("Closing serial line...")
            self.iface.close()

        if self.leds:
            self.leds.all_off()

    def enable(self):
        pass

    def disable(self):
        pass

    def restart(self, soft=True):
        pass

    def restart_hw(self):
        pass

class QuectelEG912(LTEModemInterface):
    ENABLE_PULSE_LEN = 1.5
    RESET_PULSE_LEN = 1.5

    REGEX_WAN_IP = re.compile(r"\+CGPADDR:\s*\d+,\"([^\"]*)\"")
    REGEX_LOCAL_IP = re.compile(r'(?<=inet )((?:\d+\.){3}\d+)')

    def __init__(self, nic_usb_port, led_definitions, at_port=None):
        super().__init__(nic_usb_port, led_definitions, at_port)

    def get_ip_addresses(self):

        ip = QuectelEG912.REGEX_WAN_IP.search(str(modem.query_extended_timeout("AT+CGPADDR=1\r\n", timeout=3.0)))
        if isinstance(ip, re.Match):
            ip = ip.group(1)

        ret = subprocess.check_output("ip addr show " + current_config['lte_iface'], shell=True)
        priv_ip = QuectelEG912.REGEX_LOCAL_IP.search(str(ret))
        if isinstance(priv_ip, re.Match):
            priv_ip = priv_ip.group(1)

        return ip, priv_ip

    def init(self):

        super().init()

        self.iface.read(4065)  # Dummy read to flush the input buffer
        retry_counter = 5
        while retry_counter:
            try:
                ret = self.query("AT\r\n", 10)  # Check if modem responds to AT commands
                if LTEModem.AT_OK_REPLY in ret:
                    break
            except TimeoutException:
                retry_counter = retry_counter - 1
                logger.debug("No reply from modem, retries left {}...".format(retry_counter))
                time.sleep(0.5)
                continue
            raise TimeoutException

        while LTEModem.AT_OK_REPLY not in self.query("AT+CMEE=1\r\n",
                                                     17):  # Set extended error report (see CME codes in manual)
            time.sleep(2)
        while True:  # Manage pin stuff
            try:
                chars = self.query("AT+CPIN?\r\n")
                if (b"SIM PIN" in chars):
                    if (current_config['pin'] != None):
                        self.query("AT+CPIN={}\r\n".format(current_config['pin']))
                    else:  # Prevent to inserting invalid PIN if undefined in config file
                        logger.debug("No PIN code defined in {}".format(CONFILE_PATH))
                        raise ModemErrorException(
                            "ERROR: 16")  # This is the same ERR code as generated by the LTE modem.
                    continue
                elif (b"SIM PUK" in chars):
                    logger.warning("PUK code needed! Use AT Command interface to unlock the SIM.")
                    raise ModemErrorException("ERROR: 12")  # This is the same ERR code as generated by the LTE modem.
                    break
                elif (b"READY" in chars):
                    logger.info("Pin resolved")
                    self.leds.set('sim', 'on')
                    break
                else:
                    logger.debug("UNEXPECTED DATA RECEIVED len({}) {}".format(len(chars), chars))

            except Exception as exc:
                if "ERROR: 14" in str(exc):
                    logger.info("Sim busy, waiting...")
                    continue
                else:
                    raise exc

        while True:  # Set APN
            # method = int(current_config['auth_method']) # Currently not supported for EG912Y
            current_state = self.query("AT+CGDCONT?\r\n")

            at_content = "1,{},\"0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0\",0,0,0,0,1,1".format(
                current_config['apn'])

            if (at_content.encode() in current_state):
                logger.debug("APN remains the same : " + current_config['apn'])
                break
            else:
                logger.info("Changing APN to " + at_content)

                self.query("AT+CGDCONT={}\r\n".format(at_content))

                raise NewSettinngs  # Restart is needed

        current_state = self.query("AT+QNETDEVCTL?\r\n")

        logger.debug("Current connection status QNETDEVCTL : " + str(current_state))
        if ("0,0".encode() in current_state):
            logger.debug("Enabling connection")
            current_state = self.query("AT+QNETDEVCTL=3,1,0\r\n")  # Parameter 3 = connection auto keep-on
        else:
            logger.debug("Already connected")

        ConnectionGuard.try_interface_up(current_config['lte_iface'])
        # os.system("ip link set {} up".format(current_config['lte_iface']))
        logger.info("Modem init complete")

    def resolve_error(self, exc):

        action = ACTION_NOTHING
        errcode = ERROR_CODE_UNSPEC
        logger = logging.getLogger('unipilte')
        errstring = str(exc)

        if isinstance(exc, TimeoutException):
            logger.error("Modem reply timeout")
            action = ACTION_HARDRESTART
        elif "ERROR: 10" in errstring:
            logger.error("SIM Not inserted")
            self.leds.set('sim', 'blink_slow')
            errcode = ERROR_CODE_NO_SIM
            action = ACTION_SOFTRESTART
        elif "ERROR: 13" in errstring:
            logger.error("SIM Failure")
            self.leds.set('sim', 'blink_slow')
            errcode = ERROR_CODE_SIM_ERROR
            action = ACTION_SOFTRESTART
        elif "ERROR: 14" in errstring:
            logger.info("SIM Busy")
            self.leds.set('sim', 'blink_slow')
            errcode = ERROR_CODE_SIM_ERROR
            action = ACTION_NOTHING
        elif "ERROR: 16" in errstring:
            logger.error("Incorrect or undefined password (PIN). Going to sleep...")
            self.leds.set('sim', 'blink_fast')
            errcode = ERROR_CODE_WRONG_PIN
            action = ACTION_HALT
        elif "ERROR: 12" in errstring:
            logger.error("SIM Locked, PUK needed. Going to sleep...")
            self.leds.set('sim', 'blink_fast')
            errcode = ERROR_CODE_WRONG_PIN
            action = ACTION_HALT
        elif "Modem I/O error" in errstring:
            logger.error("{}, performing HARD restart".format(errstring))
            action = ACTION_HARDRESTART
        elif "ERROR: 569" in errstring:
            logger.error("An error with code 569")
            action = ACTION_NOTHING
        else:
            logger.error("Unhandled error occured: " + errstring)
            action = ACTION_SOFTRESTART

        # Time filter - if more than THRESHOLD restarts occurs with minimal delay, daemon goes to SUSPEND
        if (action == ACTION_HARDRESTART or action == ACTION_SOFTRESTART):
            if (time.time() - LTEModem.last_rst_time) < 60.0:
                logger.info("Too rapid restart request, {} left to SUSPEND ".format(10 - LTEModem.rept_rst_cnt))
                LTEModem.rept_rst_cnt = LTEModem.rept_rst_cnt + 1

            if LTEModem.rept_rst_cnt > 10:
                logger.warning("Maximum number of resets reached, going to SUSPEND".format())
                LTEModem.rept_rst_cnt = 0
                LTEModem.last_rst_time = 0
                action = ACTION_SUSPEND

            LTEModem.last_rst_time = time.time()

        return action, errcode
    
    
    def restart(self, soft=True):

        if soft == True:  # SW restart

            self.leds.all_blink()

            chars = self.query_extended_timeout("AT+CFUN=1,1\r\n", 3.0, 15)
            self.iface.close()

        else:  # HW restart
            try:
                self.restart_hw()
            except BaseException as E:
                logger.error("MODEM HW RESET ERROR: {}".format(str(E)))
                print("Exc: " + str(E))
                return

        while True:
            logger.debug("Waiting for modem disconnect...")
            try:
                stat.S_ISCHR(os.stat(self.at_port_path).st_mode)
            except:
                break
            time.sleep(0.3)

        logger.debug("Disconnected")

        while True:
            logger.debug("Waiting for modem reconnect...")
            try:
                stat.S_ISCHR(os.stat(self.at_port_path).st_mode)
            except:
                time.sleep(0.3)
                continue
            break

        logger.debug("Reconnected")

        self.iface.open()
        self.iface.flush()
        self.iface.reset_output_buffer()
        self.iface.reset_input_buffer()
        while True:
            self.iface.write("AT\r\n".encode())
            if (len(chars) > 0):
                break

        logger.info("RESTART complete OK".format(len(chars), chars))

        self.leds.all_off()

        time.sleep(random.random() * 10 + 5.0)

        if not current_config['lte_iface']:
            current_config['lte_iface'] = get_lte_iface_name(self.nic_usb_port)

class ID_LTE_1(QuectelEG912):
    
    def __init__(self, nic_usb_port, rfkill_enable, rfkill_reset, at_port=None):
        super().__init__(nic_usb_port, PLATFORM_SPEC["ID-LTE-1"]['leds'], at_port)
        self.rfkill_enable = rfkill_enable
        self.rfkill_reset = rfkill_reset
        
        
    def enable(self):
        with open(self.rfkill_enable + "/soft", "w") as f:
            f.write("0")
        

    def hw_restart(self):
        with open(self.rfkill_reset + "/soft", "w") as f:
            f.write("1")
                    


class I_1006_3(QuectelEG912):
    
    def __init__(self, enable_pin, reset_pin, nic_usb_port, at_port=None):
        super().__init__(nic_usb_port, PLATFORM_SPEC["Patron"]['leds'], at_port)
        
        self.enable_pin = enable_pin
        self.reset_pin = reset_pin
        self.enable_pulse = QuectelEG912.ENABLE_PULSE_LEN
        self.reset_pulse = QuectelEG912.RESET_PULSE_LEN

    def enable(self, restart=False):
        global logger

        toggle_pin = self.enable_pin
        oper_str = "Starting" if restart is False else "Restarting"

        if (restart is True):
            toggle_pin = self.reset_pin

        delay = self.enable_pulse if restart is False else self.reset_pulse

        try:
            ret = os.path.exists('/sys/class/gpio/gpio' + str(toggle_pin))
            if ret == False:  # If not exported yet...
                with open('/sys/class/gpio/export', 'w') as f:  # Export it !
                    f.write(str(toggle_pin))
                    f.close()

            with open('/sys/class/gpio/gpio' + str(toggle_pin) + '/direction', 'w') as f:
                f.write('out')  # Set GPIO as output
                f.close()

            with open('/sys/class/gpio/gpio' + str(toggle_pin) + '/value', 'w') as f:
                if logger:
                    logger.debug("{} LTE stage 1".format(oper_str))
                else:
                    print("{} LTE stage 1".format(oper_str))
                f.write("1")  # There is an inverting transistor before the RESET_N pin of the modem !
                f.flush()
                time.sleep(delay)  # This time should be 2.3 - 15 sec according the datasheet
                if logger:
                    logger.debug("{} LTE stage 2".format(oper_str))
                else:
                    print("{} LTE stage 2".format(oper_str))
                f.write("0")  # Back to LOG 1
                f.flush()
                f.close()

        except BaseException as E:
            # logger.error("MODEM HW RESET ERROR: {}".format(str(E)))
            if logger:
                logger.error("Enable failed - {}".format(str(E)))
            else:
                print("Exc: " + str(E))
                print(traceback.format_exc())

    def restart_hw(self): 
        self.enable(restart=True)

class ModemFactory(object):

    # Try to read current unipi platform based on CPU type
    # TODO - use unipi-id (always present since Debian 11) instead
    @staticmethod
    def get_platform():
        model = "Generic"
        platform = "N/A"

        MODEL_ID_FILE = "/sys/firmware/devicetree/base/unipi-model"
        PROCINFO_FILE = "/proc/cpuinfo"

        if os.path.isfile(MODEL_ID_FILE):
            with open(MODEL_ID_FILE, 'rb') as f:
                model = f.read()
                printables = {ord(c) for c in string.printable}
                model = "".join(chr(c) for c in model if c in printables)
            if len(model) == 4 and model[3] == '7':
                platform = "Patron"
            elif len(model) == 4 and model[3] == '5':
                platform = "Axon"
            elif "IRIS" in model or model[0] == 'D':
                platform = "IRIS"
                model = "ID-LTE-1"
            elif any(oem_id in model for oem_id in ("ACPLC", "ACHE")):
                platform = "ACH-OEM"
                model = "ID-LTE-1"

        elif os.path.isfile(PROCINFO_FILE):
            with open(PROCINFO_FILE) as f:
                if "Allwinnersun50iw2Family" in f.read():
                    platform = "Axon"

        return platform, model

    @staticmethod
    def create_modem_instance() -> LTEModemInterface:
        # Resolve platform

        global logger

        platform, model = ModemFactory.get_platform()
        if logger is not None:
            logger.info("Running on {} series, model: {} ".format(platform, model))
        if platform == "Patron":
            return I_1006_3(PLATFORM_SPEC['Patron']['pwrkey_pin'], PLATFORM_SPEC['Patron']['reset_pin'],
                                PLATFORM_SPEC['Patron']['usb_port'])
        elif platform == "Axon":
            # return LTEModem(PLATFORM_SPEC['Axon']['usb_port'])
            return LTEModem(PLATFORM_SPEC['Axon']['restart_pin'], PLATFORM_SPEC['Axon']['usb_port'])
        elif platform == "IRIS" or platform == "ACH-OEM":
                
            return ID_LTE_1("/run/unipi-plc/by-sys/lte.1/interface",
                            PLATFORM_SPEC['ID-LTE-1']['rfkill_enable'],
                            PLATFORM_SPEC['ID-LTE-1']['rfkill_reset'],
                            "/run/unipi-plc/by-sys/lte.1/tty1")
        else:
            print("UNSUPPORTED PLATFORM {} {}".format(platform, model))
            if logger is not None:
                logger.error("Running on unsupported {} series, model: {} ".format(platform, model))
            sys.exit(0)


class LTEModem(object):
    rept_rst_cnt = 0
    last_rst_time = time.time()

    AT_OK_REPLY = b"OK"

    REGEX_CSQ = re.compile(r"\+CSQ:\s*(\d+),(\d+)")
    REGEX_COP = re.compile(r"\+COPS:\s*\d+,\d?,\"(.*)\",(\d?)")

    def __init__(self, reset_pin, nic_usb_port, at_port_path=None, retry_counter=5):
        self.at_port_path = MODEM_PORT if at_port_path is None else at_port_path

        self.reset_pin = reset_pin
        self.nic_usb_port = nic_usb_port

        self.leds = None
        self.iface = None

        if not self.leds:
            self.leds = LedServer()
            self.leds.all_on()

    def __del__(self):
        print("Modem ending...")
        if self.iface:
            print("Closing serial line...")
            self.iface.close()

        if self.leds:
            self.leds.all_off()

    def __iface_init(self, port):
        if not stat.S_ISCHR(os.stat(port).st_mode):  # Check if modem virtual serial interface exists
            raise ModemErrorException("Modem serial port {} is not a valid character device".format(port))
        if hasattr(self.iface, 'open'):
            self.iface.close()

        try:
            self.iface = serial.Serial(port, 115200, timeout=0.3, parity=serial.PARITY_NONE, rtscts=0)
        except Exception as E:
            raise ModemErrorException("Cannot open virtual serial port {}".format(USB_PORT))

        current_config['lte_iface'] = get_lte_iface_name(self.nic_usb_port)
        if not current_config['lte_iface']:
            raise ModemErrorException("No network interface found on USB port {}:{}".format(port, USB_PORT))

    def init(self):

        retry_counter = 5
        delay = 5

        if self.leds:
            self.leds.all_blink(fast=False)

        while retry_counter > 0:
            try:
                self.iface = None
                if not self.leds:
                    self.leds = LedServer()
                    self.leds.all_on()
                self.__iface_init(MODEM_PORT)
                self.leds.all_off()
                break
            except Exception as E:
                retry_counter = retry_counter - 1
                if retry_counter < 2:
                    logger.info("Modem init retry with HW restart...{} attempts left".format(retry_counter))
                    # self.restart(soft = False)
                    LTEModem.restart_hw()
                    delay = 10  # Extend delay to wait for modem boot
                else:
                    logger.info("Modem init retry...{} attempts left".format(retry_counter))

                sleep(random.random() * 10 + delay)
                exc = E

                continue

        if retry_counter == 0:
            raise ModemCreateInstanceException(str(exc))

        self.iface.read(4065)  # Dummy read to flush the input buffer
        retry_counter = 5
        while retry_counter:
            try:
                ret = self.query("AT\r\n", 10)  # Check if modem responds to AT commands
                if LTEModem.AT_OK_REPLY in ret:
                    break
            except TimeoutException:
                retry_counter = retry_counter - 1
                logger.debug("No reply from modem, retries left {}...".format(retry_counter))
                time.sleep(0.5)
                continue
            raise TimeoutException

        while LTEModem.AT_OK_REPLY not in self.query("AT+CMEE=1\r\n",
                                                     17):  # Set extended error report (see CME codes in manual)
            time.sleep(2)
        while True:  # Manage pin stuff

            chars = self.query("AT+CPIN?\r\n")
            if (b"SIM PIN" in chars):
                if (current_config['pin'] != None):
                    self.query("AT+CPIN={}\r\n".format(current_config['pin']))
                else:  # Prevent to inserting invalid PIN if undefined in config file
                    logger.debug("No PIN code defined in {}".format(CONFILE_PATH))
                    raise ModemErrorException("ERROR: 16")  # This is the same ERR code as generated by the LTE modem.
                continue
            elif (b"READY" in chars):
                logger.info("Pin resolved")
                self.leds.set('sim', 'on')
                break
            else:
                logger.debug("UNEXPECTED DATA RECEIVED len({}) {}".format(len(chars), chars))

        chars = self.query("AT+UUSBCONF?\r\n")  # Set appropriate USB mode (if needed)
        if ((b"0x1143" in chars and current_config['usbmode'] == "2") or
                (b"0x1146" in chars and current_config['usbmode'] == "3")):
            logger.debug("USB profile remains the same")
        else:
            logger.info("Changing USB mode to " + str(current_config['usbmode']))
            self.query("AT+UUSBCONF = {}\r\n".format(current_config['usbmode']))
            raise NewSettinngs  # Restart is needed

        while True:  # Set APN
            method = int(current_config['auth_method'])
            current_state = self.query("AT+UCGDFLT?\r\n")
            at_content = current_config['apn']

            if method in {1, 2}:  # With auth. - no name/password required
                at_content = at_content + f",0,0,0,0,0,0,0,0,0,1,0,0,1,0,0,0,0,{current_config['auth_method']},{current_config['auth_user']},{current_config['auth_pass']}"
            else:
                at_content = at_content + ",0,0,0,0,0,0,0,0,0,1,0,0,1,0,0,0,0,0"

            if ((method == 0 and at_content.encode() in current_state) or (
                    current_config['_was_changed'] == False and method in {1, 2})):
                logger.debug("APN remains the same : " + at_content)
                break
            else:
                logger.info("Changing APN to " + at_content)
                self.query("AT+UCGDFLT=1,{0}\r\n".format(at_content))

                raise NewSettinngs  # Restart is needed

        ConnectionGuard.try_interface_up(current_config['lte_iface'])
        # os.system("ip link set {} up".format(current_config['lte_iface']))
        logger.info("Modem init complete")

    def query(self, query, bytes_expected=569):
        chars = ""
        retry_cnt = 3

        while retry_cnt > 0:
            try:
                self.iface.flushOutput()  # Flush da buffer
                self.iface.flushInput()  # Flush da buffer
                # time.sleep(0.1)
                self.iface.write(query.encode())  # Write op. is blocking - may take longer !
                chars = self.iface.read(bytes_expected)  # Read with previously defined timeout
            except Exception as e:  # May occur when other app uses modem's serial line
                raise ModemErrorException("Modem I/O error " + str(e))
            # logger.debug("RECEIVED {} bytes: {}".format(len(chars),chars))
            if b"ERROR" in chars:
                raise ModemErrorException(chars)
            if len(chars) > 0:  # No reply received
                return chars

            retry_cnt = retry_cnt - 1
            if len(query) > 4:  # Supress AT\r\n polling during startup sequence
                logger.info("More attempts for " + str(query).strip('\r\n'))

        raise TimeoutException

    def query_extended_timeout(self, query, timeout, bytes_expected=569):

        ret = ""
        timeout_bckp = self.iface.timeout
        self.iface.timeout = timeout
        try:
            ret = self.query(query, bytes_expected)
        finally:
            self.iface.timeout = timeout_bckp
        return ret

    @staticmethod
    def restart_hw():  # Hardware restart by RST pin toggling. Static function - not tied with any instance.
        global logger
        restart_pin = PLATFORM_SPEC["Axon"]["restart_pin"]
        try:
            ret = os.path.exists('/sys/class/gpio/gpio' + str(restart_pin))
            if ret == False:  # If not exported yet...
                with open('/sys/class/gpio/export', 'w') as f:  # Export it !
                    f.write(str(restart_pin))
                    f.close()

            with open('/sys/class/gpio/gpio' + str(restart_pin) + '/direction', 'w') as f:
                f.write('out')  # Set GPIO as output
                f.close()

            with open('/sys/class/gpio/gpio' + str(restart_pin) + '/value', 'w') as f:
                if logger:
                    logger.debug("Restarting LTE stage 1")
                else:
                    print("Restarting LTE stage 1")
                f.write("1")  # There is an inverting transistor before the RESET_N pin of the modem !
                f.flush()
                time.sleep(3)  # This time should be 2.3 - 15 sec according the datasheet
                if logger:
                    logger.debug("Restarting LTE stage 2")
                else:
                    print("Restarting LTE stage 2")
                f.write("0")  # Back to LOG 1
                f.flush()
                f.close()

        except BaseException as E:
            # logger.error("MODEM HW RESET ERROR: {}".format(str(E)))
            if logger:
                logger.error("Hard reset failed - {}".format(str(E)))
            else:
                print("Exc: " + str(E))

    def restart(self, soft=True):

        if soft == True:  # SW restart

            self.leds.all_blink()

            chars = self.query_extended_timeout("AT+CFUN=16\r\n", 3.0, 15)
            self.iface.close()

        else:  # HW restart
            try:
                LTEModem.restart_hw()
            except BaseException as E:
                logger.error("MODEM HW RESET ERROR: {}".format(str(E)))
                print("Exc: " + str(E))
                return

        while True:
            logger.debug("Waiting for modem disconnect...")
            try:
                stat.S_ISCHR(os.stat(MODEM_PORT).st_mode)
            except:
                break
            time.sleep(0.3)

        logger.debug("Disconnected")

        while True:
            logger.debug("Waiting for modem reconnect...")
            try:
                stat.S_ISCHR(os.stat(MODEM_PORT).st_mode)
            except:
                time.sleep(0.3)
                continue
            break

        logger.debug("Reconnected")

        self.iface.open()
        self.iface.flush()
        self.iface.reset_output_buffer()
        self.iface.reset_input_buffer()
        while True:
            self.iface.write("AT\r\n".encode())
            if (len(chars) > 0):
                break

        logger.debug("RESTART complete ".format(len(chars), chars))

        self.leds.all_off()

        time.sleep(random.random() * 10 + 5.0)

        if not current_config['lte_iface']:
            current_config['lte_iface'] = get_lte_iface_name(USB_PORT)

    def resolve_error(self, exc):

        action = ACTION_NOTHING
        errcode = ERROR_CODE_UNSPEC
        logger = logging.getLogger('unipilte')
        errstring = str(exc)

        if isinstance(exc, TimeoutException):
            estr = str(exc) if len(str(exc)) != 0 else "Modem reply timeout"
            logger.error(str(exc))
            action = ACTION_HARDRESTART
        elif "ERROR: 10" in errstring:
            logger.error("SIM Not inserted")
            self.leds.set('sim', 'blink_slow')
            errcode = ERROR_CODE_NO_SIM
            action = ACTION_SOFTRESTART
        elif "ERROR: 13" in errstring:
            logger.error("SIM Failure")
            self.leds.set('sim', 'blink_slow')
            errcode = ERROR_CODE_SIM_ERROR
            action = ACTION_SOFTRESTART
        elif "ERROR: 16" in errstring:
            logger.error("Incorrect or undefined password (PIN). Going to sleep...")
            self.leds.set('sim', 'blink_fast')
            errcode = ERROR_CODE_WRONG_PIN
            action = ACTION_HALT
        elif "Modem I/O error" in errstring:
            logger.error("{}, performing HARD restart".format(errstring))
            action = ACTION_HARDRESTART
        elif "ERROR: 569" in errstring:
            logger.error("An error with code 569")
            action = ACTION_NOTHING
        else:
            logger.error("Unhandled error occured: " + errstring)
            action = ACTION_SOFTRESTART

        # Time filter - if more than THRESHOLD restarts occurs with minimal delay, daemon goes to SUSPEND
        if (action == ACTION_HARDRESTART or action == ACTION_SOFTRESTART):
            if (time.time() - LTEModem.last_rst_time) < 60.0:
                logger.info("Too rapid restart request, {} left to SUSPEND ".format(10 - LTEModem.rept_rst_cnt))
                LTEModem.rept_rst_cnt = LTEModem.rept_rst_cnt + 1

            if LTEModem.rept_rst_cnt > 10:
                logger.warning("Maximum number of resets reached, going to SUSPEND".format())
                LTEModem.rept_rst_cnt = 0
                LTEModem.last_rst_time = 0
                action = ACTION_SUSPEND

            LTEModem.last_rst_time = time.time()

        return action, errcode

    def update_led(self, is_registered, rssi, wan_state):

        if is_registered is True:  # Registered in the network
            self.leds.set('net', 'on')
        else:
            self.leds.set('net', 'off')

        if rssi == 99:
            self.leds.set('sig', 'off')
        else:
            self.leds.duty('sig', rssi)

        if current_config['ping']['autoswitch']:
            if wan_state == WAN_STATE_LTE_ON_ETH_ON:  # Connection to the internet (or other WAN network)
                self.leds.set('int', 'on')
            elif wan_state in {WAN_STATE_LTE_ON_ETH_OFF, WAN_STATE_LTE_ON}:
                self.leds.set('int', 'blink_fast')
            elif wan_state in {WAN_STATE_LTE_OFF_ETH_ON, WAN_STATE_ETH_ON}:
                self.leds.set('int', 'blink_slow')
            else:  # OFFLINE
                self.leds.set('int', 'off')
        else:  # No autoswitch
            if wan_state in {WAN_STATE_ETH_ON, WAN_STATE_LTE_ON}:
                self.leds.set('int', 'on')
            else:
                self.leds.set('int', 'off')

    def go_halt_suspend(self, errcode=ERROR_CODE_UNSPEC):

        if errcode in (ERROR_CODE_NO_SIM, ERROR_CODE_SIM_ERROR):
            self.leds.all_off()
            self.leds.set('sim', 'blink_slow')
        elif errcode == ERROR_CODE_WRONG_PIN:
            self.leds.all_off()
            self.leds.set('sim', 'blink_fast')
        else:
            self.leds.all_blink(fast=False)

    def get_operator_nettype(self):
        ret = self.query("AT+COPS?\r\n")
        res = self.REGEX_COP.search(str(ret))
        if res is None or len(res.groups()) != 2:
            # logger.info("Cannot parse operator info, received: {}".format(ret))
            return (None, None)

        return (res.group(1), res.group(2))

    def set_local_nic(self, ip, mask, dhcp_range_start, dhcp_range_end):
        ret = self.query("AT+UIPCONF=\"{}\",\"{}\",\"{}\",\"{}\"\r\n".ip,
                         mask, dhcp_range_start, dhcp_range_end)

    def get_sig_metrics(self):
        ret = self.query("AT+CSQ\r\n")
        res = self.REGEX_CSQ.search(str(ret))
        if res is None or len(res.groups()) != 2:
            logger.info("Cannot parse signal quality, received: {}".format(ret))
            return (99, 0)

        return (int(res.group(1)), int(res.group(2)))


# --------------------------------------------------------------------------
# --------------------------------------------------------------------------
# --------------------------------------------------------------------------


parser = argparse.ArgumentParser(description="2G/3G/LTE connection management daemon", epilog="unipi.technology 2019")
parser.add_argument("-r", "--reset", dest="reset", action="store_true", help="perform HW reset of the LTE module")
parser.add_argument("-d", "--delete", dest="delete", action="store_true", help="delete NIC from the system")
args = parser.parse_args()
NetworkManagerDriver.init()

if args.reset:  # Single shot reset (call the script with -r or --reset)
    modem = ModemFactory.create_modem_instance()
    modem.restart_hw()
    sys.exit(0)

if os.path.isfile(PIDFILE_PATH):
    print("LTE Daemon already running, exiting...")
    sys.exit(0)

if args.delete:  # Single shot reset (call the script with -r or --reset)
    modem = ModemFactory.create_modem_instance()
    current_config['lte_iface'] = get_lte_iface_name(modem.nic_id, create_config_file=False)
    remove_iface_confile(current_config['lte_iface'])

    del modem

    ConnectionGuard.update_resolver_file(lte_as_dns=False, dns_ip=None)
    sys.exit(0)

try:  # Daemon mode starts here

    pid = str(os.getpid())
    file = open(PIDFILE_PATH, 'w')
    file.write(pid)
    file.close()

    signal.signal(signal.SIGINT, sig_handler)  # LOG exit cause by signal in handler
    signal.signal(signal.SIGTERM, sig_handler)
    signal.signal(signal.SIGQUIT, sig_handler)

    conn_quard = None
    try:
        config = ConfigParser.ConfigParser(inline_comment_prefixes=(';',))
    except TypeError:
        config = ConfigParser.ConfigParser()  # In Python 2.x, the module is slightly different

    err_msg = ""

    if not os.path.isfile(CONFILE_PATH):  # Check the config file
        err_msg = "Config file not found on path: {}".format(CONFILE_PATH)

    confile_was_changed = was_confile_changed()

    try:
        config.read([CONFILE_PATH])
    except Exception as e:
        err_msg = "Error while parsing file {} : {}".format(CONFILE_PATH, str(e))

    log_path = getFromConfile(config, "LOG", "path", DEFAULT_LOG_PATH)
    log_level = getFromConfile(config, "LOG", "severity", DEFAULT_LOGGING_LEVEL).upper()
    log_size = getFromConfile(config, "LOG", "logfile_size", DEFAULT_LOGFILE_SIZE, integer=True)

    logger = logging.getLogger('unipilte')  # Global instance of the logger for this application

    try:
        logger.setLevel(log_level)
    except:
        logger.setLevel(DEFAULT_LOGGING_LEVEL)
        err_msg = "Invalid log severity"

    if not log_path:
        filelog_handler = logging.handlers.SysLogHandler(address='/dev/log')  # Standard rotated log
        log_formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    else:
        filelog_handler = logging.handlers.RotatingFileHandler(filename=log_path, maxBytes=log_size,
                                                               backupCount=1)  # Standard rotated log
        log_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

    filelog_handler.setFormatter(log_formatter)

    logger.addHandler(filelog_handler)
    logger.info("unipi-lte version {} from {} - Logging started..".format(DAEMON_VERSION,
                                                                          datetime.datetime.fromtimestamp(
                                                                              os.path.getmtime(
                                                                                  os.path.realpath(__file__)))))

    if err_msg:  # Handle log messages from config file parsing stage
        logger.warning(err_msg + "; USING DEFAULT VALUES")
        print(err_msg + "; USING DEFAULT VALUES")
    else:
        logger.debug("Config file {} parsed".format(CONFILE_PATH))

    current_config = {'_was_changed': confile_was_changed, 'apn': getFromConfile(config, "MAIN", "apn", None)}

    if not current_config['apn']:
        modem = ModemFactory.create_modem_instance()
        modem.leds.all_off()
        #os.system("ip link set {} down".format(current_config['lte_iface']))
        print("No APN defined in {}, stopping unipi-lte service...".format(CONFILE_PATH))
        logger.warning("No APN defined in {}, stopping unipi-lte service...".format(CONFILE_PATH))
        os._exit(0)

    current_config['auth_method'] = getFromConfile(config, "MAIN", "auth_method", 0, integer=True)
    if current_config['auth_method'] != 0:
        current_config["auth_user"] = getFromConfile(config, "MAIN", "auth_user", "")
        current_config["auth_pass"] = getFromConfile(config, "MAIN", "auth_pass", "")
    current_config['pin'] = getFromConfile(config, "MAIN", "pin", None)
    current_config['usbmode'] = getFromConfile(config, "MAIN", "usbmode", "2")
    current_config['report_period'] = getFromConfile(config, "MAIN", "report_period", 86400, integer=True)
    current_config['oneshot'] = getFromConfile(config, "MAIN", "oneshot", default=False, boolean=True)
    current_config['regtimeout'] = getFromConfile(config, "MAIN", "reg_timeout", 120, integer=True)
    current_config['suspend_timeout'] = getFromConfile(config, "MAIN", "suspend_timeout", 86400, integer=True)
    current_config['lte_autokeep_on'] = getFromConfile(config, "MAIN", "auto_keepon", default=True, boolean=True)

    current_config['lte_iface'] = None

    current_config['ping'] = {}
    current_config['ping']['active'] = getFromConfile(config, "PING", "active", default=False, boolean=True)

    current_config['ping']['timeout'] = int(getFromConfile(config, "PING", "timeout", 60, integer=True))
    current_config['ping']['period'] = int(getFromConfile(config, "PING", "period", 20))
    current_config['ping']['threshold'] = int(getFromConfile(config, "PING", "fail_threshold", 3))
    current_config['ping']['ip_lte'] = getFromConfile(config, "PING", "ip_lte", "8.8.8.8")
    current_config['ping']['autoswitch'] = getFromConfile(config, "PING", "autoswitch", default=False, boolean=True)
    current_config['ping']['eth_iface'] = getFromConfile(config, "PING", "primary_iface", "eth0")

    if current_config['ping']['autoswitch']:
        current_config['ping']['ip_eth'] = getFromConfile(config, "PING", "ip_eth", "8.8.4.4")

    conn_quard = ConnectionGuard()

    logger.debug("CURRENT SETTINGS: {} ".format(str(current_config)))

    # Regex parsing some specific AT commands replies

    REGEX_IP = re.compile(r"\+UIPADDR:\s*\d+,\"\w*\",\"([^\"]*)\"")
    REGEX_PRIV_IP = re.compile(r"\+UIPCONF:\s\"([^\"]*)\",\"([^\"]*)\"")

    # In-loop variables

    stats = Stats(current_config['report_period'], current_config['ping']['active'])
    logger.debug("Entering the main loop...")
    ping_timestep = 0
    reg_timeout = current_config['regtimeout']
    is_registered = False
    start_timestamp = 0

    reinit = True  # Init after start

    # sys.exit(0)
    modem = None
    while True:
        try:
            start_timestamp = time.time()
            if reinit:

                if conn_quard is not None:
                    conn_quard.reset_state()

                reinit = False
                wan_available = False
                # if not modem: # Try to create modem instance - working serial port needed
                if modem:
                    del modem

                modem = ModemFactory.create_modem_instance()

                modem.init()

                ping_timestep = 0
                reg_timeout = current_config['regtimeout']
                is_registered = False
                start_timestamp = 0

                # ------------------------------------------------------------------------------

                if current_config['oneshot']:
                    logger.info("One-shot mode, exiting...")
                    sys.exit(0)

            rssi, sigqual = modem.get_sig_metrics()

            operator, nettype = modem.get_operator_nettype()

            if operator is not None:  # Registered to the network, thus nettype is valid

                reg_timeout = current_config['regtimeout']  # Reset non-registered counter

                if is_registered is False:  # Registered for a first time
                    is_registered = True
                    logger.info(
                        "Registered to: {} Network type: {} RSSI: {} Sigqual: {}".format(operator, nettype, rssi,
                                                                                         sigqual))

                    # Check local IP address assigned by modems DHCP server
                    if hasattr(modem, 'get_ip_addresses'):
                        ip, priv_ip = modem.get_ip_addresses()
                        logger.info("WAN IP ADDRESS: {} INTERFACE IP ADDRESS: {}".format(ip, priv_ip))
                    else:
                        ip = REGEX_IP.search(str(modem.query_extended_timeout("AT+UIPADDR=\r\n", timeout=3.0)))
                        priv_ip = REGEX_PRIV_IP.search(
                            str(modem.query_extended_timeout("AT+UIPCONF?\r\n", timeout=3.0)))

                        if ip is not None:
                            new_net_settings = resolve_subnets(ip.group(1), priv_ip.group(1), priv_ip.group(2))
                            logger.info(
                                "WAN IP ADDRESS: {} INTERFACE IP ADDRESS: {}".format(ip.group(1), priv_ip.group(1)))
                            if new_net_settings:
                                modem.set_local_nic(new_net_settings['ip'], new_net_settings['mask'],
                                                    new_net_settings['dhcp'], new_net_settings['dhcp'])
                                raise NewSettinngs
                        else:
                            logger.warning("Cannot get WAN IP adress from the LTE module")

            else:  # Not registered to operator
                if is_registered is True:
                    logger.info("Deregistered from network operator. {} Cycles to restart".format(reg_timeout))
                is_registered = False
                reg_timeout -= 1
                logger.debug("Dereg counter: {}".format(reg_timeout))
                if reg_timeout == 0:
                    reg_timeout = current_config['regtimeout']
                    raise TimeoutException("Cannot register to network")
                modem.query("AT+CPIN?\r\n")  # Check da SIM

            logger.debug("RSSI: {0} sigqual: {1} net-type: {2}".format(rssi, sigqual, nettype))

            if not current_config['ping']['active'] or not operator:  # Manage ping
                modem.update_led(is_registered, rssi, WAN_STATE_MONITOR_OFF)
            elif ping_timestep:
                ping_timestep -= 1
                logger.debug("Next ping in " + str(ping_timestep) + " cycles")
                modem.update_led(is_registered, rssi, conn_quard.get_wan_status())

            else:
                ping_timestep = current_config['ping']['period']
                ping_result = conn_quard.ping()
                if not conn_quard.get_rtable_valid():
                    ping_timestep = 0  # If rtable is invalid, try ping again a bit earlier

                if current_config['lte_autokeep_on'] is True:
                    conn_quard.keep_iface_on()
                modem.update_led(is_registered, rssi, conn_quard.get_wan_status())
                stats.report_metrics(rssi, sigqual, conn_quard.get_wan_status(), nettype, ping_result)


            if stats.is_time_fullfilled():  # Generate report to the log
                logger.info(stats.get_summmary(do_reset=True))

            stats.report_metrics(rssi, sigqual, conn_quard.get_wan_status(), nettype, ping_result=None)
            time.sleep(10 - ((time.time() - start_timestamp) % 10))
            logger.debug("STEP")

        except ModemCreateInstanceException as e:
            logger.error("Cannot create modem instance: " + str(e))
            do_action(None, ACTION_SUSPEND)
            reinit = do_action(None, ACTION_HARDRESTART)
            continue

        except ConnectionError as e:
            logger.error("Cannot connect to WAN via LTE, restarting...")
            reinit = do_action(modem, ACTION_HARDRESTART)
            continue

        except (ModemErrorException, TimeoutException) as e:
            action, errcode = modem.resolve_error(e)
            reinit = do_action(modem, action, errcode)
            logger.debug("Continuing")
            continue

        except NewSettinngs as e:
            logger.info("Settings of the modem has been changed, restarting...")
            reinit = do_action(modem, ACTION_SOFTRESTART)
            continue

        except KeyboardInterrupt:
            led.close()
            logging.shutdown()
            print("Good bye...")

except LEDErrorException as ER:
    traceback.print_exc()
    print("Cannot init LED subsystem: " + str(ER))
    if logger is not None:
        logger.critical("LED subsystem error:" + str(ER))

except Exception as ER:
    traceback.print_exc()
    print("Expeption occured: " + str(ER))
    if logger is not None:
        logger.critical("Unhandled exception: " + str(ER))

finally:
    del(stats)
    os.unlink(PIDFILE_PATH)
